@@ -1,4 +1,5 @@
 # Copyright 2015, 2016 OpenMarket Ltd
+# Copyright 2021 The Matrix.org Foundation C.I.C.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,8 +16,9 @@
 from typing import Any, List
 
 from synapse.config.sso import SsoAttributeRequirement
+from synapse.types import JsonDict
 
-from ._base import Config, ConfigError
+from ._base import Config
 from ._util import validate_config
 
 
@@ -28,21 +30,17 @@ class CasConfig(Config):
 
     section = "cas"
 
-    def read_config(self, config, **kwargs):
+    def read_config(self, config: JsonDict, **kwargs: Any) -> None:
         cas_config = config.get("cas_config", None)
         self.cas_enabled = cas_config and cas_config.get("enabled", True)
 
         if self.cas_enabled:
             self.cas_server_url = cas_config["server_url"]
 
-            # The public baseurl is required because it is used by the redirect
-            # template.
-            public_baseurl = self.public_baseurl
-            if not public_baseurl:
-                raise ConfigError("cas_config requires a public_baseurl to be set")
-
             # TODO Update this to a _synapse URL.
+            public_baseurl = self.root.server.public_baseurl
             self.cas_service_url = public_baseurl + "_matrix/client/r0/login/cas/ticket"
+
             self.cas_displayname_attribute = cas_config.get("displayname_attribute")
             required_attributes = cas_config.get("required_attributes") or {}
             self.cas_required_attributes = _parsed_required_attributes_def(
@@ -55,37 +53,6 @@ class CasConfig(Config):
             self.cas_displayname_attribute = None
             self.cas_required_attributes = []
 
-    def generate_config_section(self, config_dir_path, server_name, **kwargs):
-        return """\
-        # Enable Central Authentication Service (CAS) for registration and login.
-        #
-        cas_config:
-          # Uncomment the following to enable authorization against a CAS server.
-          # Defaults to false.
-          #
-          #enabled: true
-
-          # The URL of the CAS authorization endpoint.
-          #
-          #server_url: "https://cas-server.com"
-
-          # The attribute of the CAS response to use as the display name.
-          #
-          # If unset, no displayname will be set.
-          #
-          #displayname_attribute: name
-
-          # It is possible to configure Synapse to only allow logins if CAS attributes
-          # match particular values. All of the keys in the mapping below must exist
-          # and the values must match the given value. Alternately if the given value
-          # is None then any value is allowed (the attribute just must exist).
-          # All of the listed attributes must match for the login to be permitted.
-          #
-          #required_attributes:
-          #  userGroup: "staff"
-          #  department: None
-        """
-
 
 # CAS uses a legacy required attributes mapping, not the one provided by
 # SsoAttributeRequirement.
